/*
 * kgr_patch_bsc1195908
 *
 * Fix for CVE-2022-0492, bsc#1195908
 *
 *  Upstream commit:
 *  24f600856418 ("cgroup-v1: Require capabilities to set release_agent")
 *
 *  SLE12-SP3 commit:
 *  50d93c2591da9e4228629449979e6fc8dc453c5b
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  25a96a7112ab07894c6387ebb67592db9ee16315
 *
 *  SLE15-SP2 and -SP3 commit:
 *  413d689eb818bd67df51d4da49daa164e47bc89d
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Marcos Paulo de Souza <mpdesouza@suse.com>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/cgroup.h>

/* klp-ccp: from include/linux/sched.h */
static rwlock_t (*kgre_tasklist_lock);

static struct dentry *(*kgre_kernfs_mount_ns)(struct file_system_type *fs_type, int flags,
			       struct kernfs_root *root, unsigned long magic,
			       bool *new_sb_created, const void *ns);

static struct super_block *(*kgre_kernfs_pin_sb)(struct kernfs_root *root, const void *ns);

static inline struct dentry *
kgrr_kernfs_mount(struct file_system_type *fs_type, int flags,
		struct kernfs_root *root, unsigned long magic,
		bool *new_sb_created)
{
	return (*kgre_kernfs_mount_ns)(fs_type, flags, root,
				magic, new_sb_created, NULL);
}

/* klp-ccp: from include/linux/cgroup.h */
#ifdef CONFIG_CGROUPS

static struct css_set (*kgre_init_css_set);

#else /* !CONFIG_CGROUPS */
#error "klp-ccp: non-taken branch"
#endif /* !CONFIG_CGROUPS */

/* klp-ccp: from kernel/cgroup.c */
#include <linux/cred.h>
#include <linux/errno.h>
#include <linux/init_task.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/magic.h>
#include <linux/mm.h>
#include <linux/mutex.h>
#include <linux/rcupdate.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/percpu-rwsem.h>
#include <linux/string.h>
#include <linux/cgroupstats.h>
#include <linux/pid_namespace.h>
#include <linux/idr.h>
#include <linux/delay.h>
#include <linux/atomic.h>


static struct mutex (*kgre_cgroup_mutex);
static spinlock_t (*kgre_css_set_lock);

static spinlock_t (*kgre_release_agent_path_lock);

static struct cgroup_subsys *(*kgre_cgroup_subsys)[];

extern struct cgroup_root cgrp_dfl_root;

extern typeof(cgrp_dfl_root) cgrp_dfl_root;

static bool (*kgre_cgrp_dfl_root_visible);

static struct list_head (*kgre_cgroup_roots);

#define kgrr_for_each_root(root)                                             \
        list_for_each_entry((root), kgre_cgroup_roots, root_list)

#define kgrr_for_each_subsys(ss, ssid)                                       \
        for ((ssid) = 0; (ssid) < CGROUP_SUBSYS_COUNT &&                \
             (((ss) = (*kgre_cgroup_subsys)[ssid]) || true); (ssid)++)

static int (*kgre_rebind_subsystems)(struct cgroup_root *dst_root,
			     unsigned long ss_mask);

static void (*kgre_cgroup_get)(struct cgroup *cgrp);

static void cgroup_put(struct cgroup *cgrp)
{
	css_put(&cgrp->self);
}

struct cgrp_cset_link {
	/* the cgroup and css_set this link associates */
	struct cgroup		*cgrp;
	struct css_set		*cset;

	/* list of cgrp_cset_links anchored at cgrp->cset_links */
	struct list_head	cset_link;

	/* list of cgrp_cset_links anchored at css_set->cgrp_links */
	struct list_head	cgrp_link;
};

static bool kgrr_css_set_populated(struct css_set *cset)
{
	lockdep_assert_held(&(*kgre_css_set_lock));

	return !list_empty(&cset->tasks) || !list_empty(&cset->mg_tasks);
}

static void (*kgre_cgroup_update_populated)(struct cgroup *cgrp, bool populated);

static void kgrr_css_set_update_populated(struct css_set *cset, bool populated)
{
	struct cgrp_cset_link *link;

	lockdep_assert_held(&(*kgre_css_set_lock));

	list_for_each_entry(link, &cset->cgrp_links, cgrp_link)
		(*kgre_cgroup_update_populated)(link->cgrp, populated);
}

static inline void get_css_set(struct css_set *cset)
{
	atomic_inc(&cset->refcount);
}

static struct cgroup_root *cgroup_root_from_kf(struct kernfs_root *kf_root)
{
	struct cgroup *root_cgrp = kf_root->kn->priv;

	return root_cgrp->root;
}

static void (*kgre_cgroup_free_root)(struct cgroup_root *root);

static void (*kgre_cgroup_kn_unlock)(struct kernfs_node *kn);

static struct cgroup *(*kgre_cgroup_kn_lock_live)(struct kernfs_node *kn);

static int (*kgre_rebind_subsystems)(struct cgroup_root *dst_root,
			     unsigned long ss_mask);

struct cgroup_sb_opts {
	unsigned long subsys_mask;
	unsigned int flags;
	char *release_agent;
	bool cpuset_clone_children;
	char *name;
	/* User explicitly requested empty subsystem */
	bool none;
};

static int (*kgre_parse_cgroupfs_options)(char *data, struct cgroup_sb_opts *opts);

int kgrp_cgroup_remount(struct kernfs_root *kf_root, int *flags, char *data)
{
	int ret = 0;
	struct cgroup_root *root = cgroup_root_from_kf(kf_root);
	struct cgroup_sb_opts opts;
	unsigned long added_mask, removed_mask;

	if (root == &cgrp_dfl_root) {
		pr_err("remount is not allowed\n");
		return -EINVAL;
	}

	mutex_lock(&(*kgre_cgroup_mutex));

	/* See what subsystems are wanted */
	ret = (*kgre_parse_cgroupfs_options)(data, &opts);
	if (ret)
		goto out_unlock;

	if (opts.subsys_mask != root->subsys_mask || opts.release_agent)
		pr_warn("option changes via remount are deprecated (pid=%d comm=%s)\n",
			task_tgid_nr(current), current->comm);

	/* See cgroup_mount release_agent handling */
	if (opts.release_agent && !capable(CAP_SYS_ADMIN)) {
		ret = -EINVAL;
		goto out_unlock;
	}

	added_mask = opts.subsys_mask & ~root->subsys_mask;
	removed_mask = root->subsys_mask & ~opts.subsys_mask;

	/* Don't allow flags or name to change at remount */
	if ((opts.flags ^ root->flags) ||
	    (opts.name && strcmp(opts.name, root->name))) {
		pr_err("option or name mismatch, new: 0x%x \"%s\", old: 0x%x \"%s\"\n",
		       opts.flags, opts.name ?: "", root->flags, root->name);
		ret = -EINVAL;
		goto out_unlock;
	}

	/* remounting is not allowed for populated hierarchies */
	if (!list_empty(&root->cgrp.self.children)) {
		ret = -EBUSY;
		goto out_unlock;
	}

	ret = (*kgre_rebind_subsystems)(root, added_mask);
	if (ret)
		goto out_unlock;

	(*kgre_rebind_subsystems)(&cgrp_dfl_root, removed_mask);

	if (opts.release_agent) {
		spin_lock(&(*kgre_release_agent_path_lock));
		strcpy(root->release_agent_path, opts.release_agent);
		spin_unlock(&(*kgre_release_agent_path_lock));
	}
 out_unlock:
	kfree(opts.release_agent);
	kfree(opts.name);
	mutex_unlock(&(*kgre_cgroup_mutex));
	return ret;
}

static bool (*kgre_use_task_css_set_links) __read_mostly;

static void kgrr_cgroup_enable_task_cg_lists(void)
{
	struct task_struct *p, *g;

	spin_lock_bh(&(*kgre_css_set_lock));

	if ((*kgre_use_task_css_set_links))
		goto out_unlock;

	(*kgre_use_task_css_set_links) = true;

	/*
	 * We need tasklist_lock because RCU is not safe against
	 * while_each_thread(). Besides, a forking task that has passed
	 * cgroup_post_fork() without seeing use_task_css_set_links = 1
	 * is not guaranteed to have its child immediately visible in the
	 * tasklist if we walk through it with RCU.
	 */
	read_lock(&(*kgre_tasklist_lock));
	do_each_thread(g, p) {
		WARN_ON_ONCE(!list_empty(&p->cg_list) ||
			     task_css_set(p) != &(*kgre_init_css_set));

		/*
		 * We should check if the process is exiting, otherwise
		 * it will race with cgroup_exit() in that the list
		 * entry won't be deleted though the process has exited.
		 * Do it while holding siglock so that we don't end up
		 * racing against cgroup_exit().
		 */
		spin_lock_irq(&p->sighand->siglock);
		if (!(p->flags & PF_EXITING)) {
			struct css_set *cset = task_css_set(p);

			if (!kgrr_css_set_populated(cset))
				kgrr_css_set_update_populated(cset, true);
			list_add_tail(&p->cg_list, &cset->tasks);
			get_css_set(cset);
		}
		spin_unlock_irq(&p->sighand->siglock);
	} while_each_thread(g, p);
	read_unlock(&(*kgre_tasklist_lock));
out_unlock:
	spin_unlock_bh(&(*kgre_css_set_lock));
}

static void (*kgre_init_cgroup_root)(struct cgroup_root *root,
			     struct cgroup_sb_opts *opts);

static int (*kgre_cgroup_setup_root)(struct cgroup_root *root, unsigned long ss_mask);

struct dentry *kgrp_cgroup_mount(struct file_system_type *fs_type,
			 int flags, const char *unused_dev_name,
			 void *data)
{
	struct super_block *pinned_sb = NULL;
	struct cgroup_subsys *ss;
	struct cgroup_root *root;
	struct cgroup_sb_opts opts;
	struct dentry *dentry;
	int ret;
	int i;
	bool new_sb;

	/*
	 * The first time anyone tries to mount a cgroup, enable the list
	 * linking each css_set to its tasks and fix up all existing tasks.
	 */
	if (!(*kgre_use_task_css_set_links))
		kgrr_cgroup_enable_task_cg_lists();

	mutex_lock(&(*kgre_cgroup_mutex));

	/* First find the desired set of subsystems */
	ret = (*kgre_parse_cgroupfs_options)(data, &opts);
	if (ret)
		goto out_unlock;

	/* look for a matching existing root */
	if (opts.flags & CGRP_ROOT_SANE_BEHAVIOR) {
		(*kgre_cgrp_dfl_root_visible) = true;
		root = &cgrp_dfl_root;
		(*kgre_cgroup_get)(&root->cgrp);
		ret = 0;
		goto out_unlock;
	}

	/*
	 * Destruction of cgroup root is asynchronous, so subsystems may
	 * still be dying after the previous unmount.  Let's drain the
	 * dying subsystems.  We just need to ensure that the ones
	 * unmounted previously finish dying and don't care about new ones
	 * starting.  Testing ref liveliness is good enough.
	 */
	kgrr_for_each_subsys(ss, i) {
		if (!(opts.subsys_mask & (1 << i)) ||
		    ss->root == &cgrp_dfl_root)
			continue;

		if (!percpu_ref_tryget_live(&ss->root->cgrp.self.refcnt)) {
			mutex_unlock(&(*kgre_cgroup_mutex));
			msleep(10);
			ret = restart_syscall();
			goto out_free;
		}
		cgroup_put(&ss->root->cgrp);
	}

	kgrr_for_each_root(root) {
		bool name_match = false;

		if (root == &cgrp_dfl_root)
			continue;

		/*
		 * If we asked for a name then it must match.  Also, if
		 * name matches but sybsys_mask doesn't, we should fail.
		 * Remember whether name matched.
		 */
		if (opts.name) {
			if (strcmp(opts.name, root->name))
				continue;
			name_match = true;
		}

		/*
		 * If we asked for subsystems (or explicitly for no
		 * subsystems) then they must match.
		 */
		if ((opts.subsys_mask || opts.none) &&
		    (opts.subsys_mask != root->subsys_mask)) {
			if (!name_match)
				continue;
			ret = -EBUSY;
			goto out_unlock;
		}

		if (root->flags ^ opts.flags)
			pr_warn("new mount options do not match the existing superblock, will be ignored\n");

		/*
		 * We want to reuse @root whose lifetime is governed by its
		 * ->cgrp.  Let's check whether @root is alive and keep it
		 * that way.  As cgroup_kill_sb() can happen anytime, we
		 * want to block it by pinning the sb so that @root doesn't
		 * get killed before mount is complete.
		 *
		 * With the sb pinned, tryget_live can reliably indicate
		 * whether @root can be reused.  If it's being killed,
		 * drain it.  We can use wait_queue for the wait but this
		 * path is super cold.  Let's just sleep a bit and retry.
		 */
		pinned_sb = (*kgre_kernfs_pin_sb)(root->kf_root, NULL);
		if (IS_ERR(pinned_sb) ||
		    !percpu_ref_tryget_live(&root->cgrp.self.refcnt)) {
			mutex_unlock(&(*kgre_cgroup_mutex));
			if (!IS_ERR_OR_NULL(pinned_sb))
				deactivate_super(pinned_sb);
			msleep(10);
			ret = restart_syscall();
			goto out_free;
		}

		ret = 0;
		goto out_unlock;
	}

	/*
	 * No such thing, create a new one.  name= matching without subsys
	 * specification is allowed for already existing hierarchies but we
	 * can't create new one without subsys specification.
	 */
	if (!opts.subsys_mask && !opts.none) {
		ret = -EINVAL;
		goto out_unlock;
	}

	/*
	 * Release agent gets called with all capabilities,
	 * require capabilities to set release agent.
	 */
	if (opts.release_agent && !capable(CAP_SYS_ADMIN)) {
		ret = -EINVAL;
		goto out_unlock;
	}

	root = kzalloc(sizeof(*root), GFP_KERNEL);
	if (!root) {
		ret = -ENOMEM;
		goto out_unlock;
	}

	(*kgre_init_cgroup_root)(root, &opts);

	ret = (*kgre_cgroup_setup_root)(root, opts.subsys_mask);
	if (ret)
		(*kgre_cgroup_free_root)(root);

out_unlock:
	mutex_unlock(&(*kgre_cgroup_mutex));
out_free:
	kfree(opts.release_agent);
	kfree(opts.name);

	if (ret)
		return ERR_PTR(ret);

	dentry = kgrr_kernfs_mount(fs_type, flags, root->kf_root,
				CGROUP_SUPER_MAGIC, &new_sb);
	if (IS_ERR(dentry) || !new_sb)
		cgroup_put(&root->cgrp);

	/*
	 * If @pinned_sb, we're reusing an existing root and holding an
	 * extra ref on its sb.  Mount is complete.  Put the extra ref.
	 */
	if (pinned_sb) {
		WARN_ON(new_sb);
		deactivate_super(pinned_sb);
	}

	return dentry;
}

ssize_t kgrp_cgroup_release_agent_write(struct kernfs_open_file *of,
					  char *buf, size_t nbytes, loff_t off)
{
	struct cgroup *cgrp;

	BUILD_BUG_ON(sizeof(cgrp->root->release_agent_path) < PATH_MAX);

	/*
	 * Release agent gets called with all capabilities,
	 * require capabilities to set release agent.
	 */
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	cgrp = (*kgre_cgroup_kn_lock_live)(of->kn);
	if (!cgrp)
		return -ENODEV;
	spin_lock(&(*kgre_release_agent_path_lock));
	strlcpy(cgrp->root->release_agent_path, strstrip(buf),
		sizeof(cgrp->root->release_agent_path));
	spin_unlock(&(*kgre_release_agent_path_lock));
	(*kgre_cgroup_kn_unlock)(of->kn);
	return nbytes;
}




#include <linux/module.h>
#include "kgr_patch_bsc1195908.h"
#include "../kallsyms_relocs.h"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "cgroup_free_root", (void *)&kgre_cgroup_free_root },
	{ "cgroup_get", (void *)&kgre_cgroup_get },
	{ "cgroup_kn_lock_live", (void *)&kgre_cgroup_kn_lock_live },
	{ "cgroup_kn_unlock", (void *)&kgre_cgroup_kn_unlock },
	{ "cgroup_setup_root", (void *)&kgre_cgroup_setup_root },
	{ "cgroup_update_populated", (void *)&kgre_cgroup_update_populated },
	{ "init_cgroup_root", (void *)&kgre_init_cgroup_root },
	{ "kernfs_mount_ns", (void *)&kgre_kernfs_mount_ns },
	{ "kernfs_pin_sb", (void *)&kgre_kernfs_pin_sb },
	{ "parse_cgroupfs_options", (void *)&kgre_parse_cgroupfs_options },
	{ "rebind_subsystems", (void *)&kgre_rebind_subsystems },
	{ "cgroup_mutex", (void *)&kgre_cgroup_mutex },
	{ "cgroup_roots", (void *)&kgre_cgroup_roots },
	{ "cgroup_subsys", (void *)&kgre_cgroup_subsys },
	{ "cgrp_dfl_root_visible", (void *)&kgre_cgrp_dfl_root_visible },
	{ "css_set_lock", (void *)&kgre_css_set_lock },
	{ "init_css_set", (void *)&kgre_init_css_set },
	{ "release_agent_path_lock", (void *)&kgre_release_agent_path_lock },
	{ "tasklist_lock", (void *)&kgre_tasklist_lock },
	{ "use_task_css_set_links", (void *)&kgre_use_task_css_set_links },
};

int kgr_patch_bsc1195908_init(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}
